#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Nov  5 19:03:37 2020

@author: diego
"""

import numpy as np
import matplotlib.pyplot as plt

#Ejemplos obtenidos de:
#    https://docs.scipy.org/doc/scipy/reference/optimize.html
#    https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.least_squares.html
#    https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.curve_fit.html
#    https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.root_scalar.html


# Ajuste de funciones genéricas +++++++++++++++++++++++++++++++++++++++++++++++

# Definimos nuestra función de prueba (ver ejemplo de derivadas de la semana pasada)


def FunTest2(x,a0,a1,a2):
    return a0*np.sin(a1*x**a2)*x

# un conjunto de puntos para graficar la curva suave
x = np.linspace(0,4,401)
y = FunTest2(x,2,2*np.pi,0.5)

# genero unos pocos datos con ruido
x2 = np.linspace(0,4,41)
y2 = FunTest2(x2,2,2*np.pi,0.5)

import datetime
fecha = datetime.datetime.now()
rng = np.random.default_rng(fecha.microsecond)      # Gracias a Gualberto 
y_noise = rng.random(x2.size)*1 - 0.5

y2 = y2 + y_noise

plt.plot(x,y,x2,y2,'o')
plt.grid()
plt.xlabel('x')
plt.ylabel('y')
plt.legend(['Función','Datos con ruido'], loc='lower left')
plt.show()

# importo la función de ajuste de funciones
from scipy.optimize import curve_fit

popt, pcov = curve_fit(FunTest2, x2, y2)            # se pide un ajuste sin restricciones

plt.plot(x,y,x2,y2,'o',x,FunTest2(x, *popt),'r--')
plt.grid()
plt.xlabel('x')
plt.ylabel('y')
plt.legend(['Función','Datos con ruido','Ajuste 1'], loc='lower left')
plt.show()

print(popt)             # Vemos que el resultado no es bueno

# intentamos un ajuste con restricciones: 0<a0<inf, 4<a1<inf, 0<a2<0.8
popt2, pcov2 = curve_fit(FunTest2, x2, y2, bounds=([0, 4, 0], [np.inf, np.inf, 0.8]))            

plt.plot(x,y,x2,y2,'o',x,FunTest2(x, *popt),'r--',x,FunTest2(x, *popt2),'k-.')
plt.grid()
plt.xlabel('x')
plt.ylabel('y')
plt.legend(['Función','Datos con ruido','Ajuste 1','Ajuste 2'], loc='lower left')
plt.show()

print(popt2)      # Notar que el resultado depende del rango de restricciones

# por último, intentamos un ajuste con un valor semilla conveniente
popt3, pcov3 = curve_fit(FunTest2, x2, y2, [1.7, 5.88, 0.39])            

plt.plot(x,y,x2,y2,'o',x,FunTest2(x, *popt),'r--',x,FunTest2(x, *popt3),'b:')
plt.grid()
plt.xlabel('x')
plt.ylabel('y')
plt.legend(['Función','Datos con ruido','Ajuste 1','Ajuste 3'], loc='lower left')
plt.show()

print(popt3)

# Raíces de funciones +++++++++++++++++++++++++++++++++++++++++++++++++++++++++

# vamos a inventarnos una función bien fea
def FunTest3(x,a0,a1,a2,a3):
    return np.sin(2*np.pi*(a0*x**3+a1*x**2+a2*x+a3))

p = np.array([0.04,-3,7,1.2])

x = np.linspace(0,2,201)

plt.plot(x,FunTest3(x,*p))
plt.grid()
plt.xlabel('x')
plt.ylabel('y')
plt.show()

# tenemos un par de caminos para seguir.  Métodos con derivadas y métodos sin derivadas.

from scipy import optimize

# método de bisección
sol_bisect = optimize.root_scalar(FunTest3, args=tuple(p), bracket=[0.75, 1], method='bisect')
print(sol_bisect.root, sol_bisect.iterations, sol_bisect.function_calls)

# método de Brent https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.brentq.html
sol_brentq = optimize.root_scalar(FunTest3, args=tuple(p), bracket=[0.75, 1], method='brentq')
print(sol_brentq.root, sol_brentq.iterations, sol_brentq.function_calls)

# ahora, si conozco la forma funcional de la derivada...

def dFTest3(x,a0,a1,a2,a3):
    return 2*np.pi*(3*a0*x**2+2*a1*x+a2)*np.cos(2*np.pi*(a0*x**3+a1*x**2+a2*x+a3))

# Método de Newton
sol_Newton = optimize.root_scalar(FunTest3, args=tuple(p), x0=0.76, fprime=dFTest3, method='newton')
print(sol_Newton.root, sol_Newton.iterations, sol_Newton.function_calls)


# Ejemplo de una función con dos mínimos.  Uno local y otro global ++++++++++++

# Definimos un polinomio que vamos a usar como prueba
x = np.linspace(0,4,41)
p = np.array([0.25,-2,11/2,-5.8,10])
y = np.polyval(p,x)

plt.plot(x,y)
plt.grid()
plt.xlabel('x')
plt.ylabel('y')
plt.show()

def FunTest(x):
    return 0.25*x**4 - 2*x**3 + 11/2*x**2 - 5.8*x + 10

# Se importa la función least_squares de el subpaquete optimize de scipy
from scipy.optimize import least_squares

res_LS_1 = least_squares(FunTest, 0)
print(res_LS_1.x,res_LS_1.cost)

res_LS_2 = least_squares(FunTest, 4)
print(res_LS_2.x,res_LS_2.cost)

# vemos que el mínimo que encuentra depende del punto inicial. Vamos a probar otros métodos.

# method{‘trf’, ‘dogbox’, ‘lm’}, optional
#    Algorithm to perform minimization.
#            ‘trf’ : Trust Region Reflective algorithm, particularly suitable for large sparse problems with bounds. Generally robust method.
#            ‘dogbox’ : dogleg algorithm with rectangular trust regions, typical use case is small problems with bounds. Not recommended for problems with rank-deficient Jacobian.
#            ‘lm’ : Levenberg-Marquardt algorithm as implemented in MINPACK. Doesn’t handle bounds and sparse Jacobians. Usually the most efficient method for small unconstrained problems.
#    Default is ‘trf’. See Notes for more information.

res_LS_3 = least_squares(FunTest, 4, method = 'dogbox')
print(res_LS_3.x,res_LS_3.cost)

res_LS_4 = least_squares(FunTest, 4, method = 'lm')
print(res_LS_4.x,res_LS_4.cost)

# Vemos que todos fallan en encontrar el mínimo absoluto

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

# Estadística +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

# Más información en:
#    https://docs.scipy.org/doc/scipy/reference/tutorial/stats.html
#    https://docs.scipy.org/doc/scipy/reference/tutorial/stats/continuous.html

from scipy import stats
from scipy.stats import norm            # importo la distribución normal

# Dado lo espeso del tema, podemos empezar imprimiendo el documento de ayuda de la distribución
# print(stats.norm.__doc__)      # Que cada uno lo estudie individualmente

# Estudiamos la distribución de probabilidades y la probabilidad acumulada.

x = np.linspace(-5,5,1001)

pdfN = norm.pdf(x)
cdfN = norm.cdf(x)
sfN = norm.sf(x)

plt.plot(x,pdfN,x,cdfN,x,sfN)
plt.grid()
plt.xlabel('x')
plt.ylabel('y')
plt.legend(['pdf', 'cdf', 'sf'], loc='center left')
plt.title('Distribución Normal')
plt.show()

# más detalles de cada método:
# The main public methods for continuous RVs are:
#    rvs: Random Variates
#    pdf: Probability Density Function
#    cdf: Cumulative Distribution Function
#    sf: Survival Function (1-CDF)
#    ppf: Percent Point Function (Inverse of CDF)
#    isf: Inverse Survival Function (Inverse of SF)
#    stats: Return mean, variance, (Fisher’s) skew, or (Fisher’s) kurtosis
#    moment: non-central moments of the distribution

# cómo generamos números aleatorios dentro de esa distribución

y = norm.rvs(size=1000)

# Análisis de la muestra (sabemos que la generamos a partir de una distribución normal)

print(y.min())
print(y.max())
print(y.mean())
print(y.var())

print(stats.describe(y))

# si queremos generar distribuciones con otras medias y desviaciones estándar

y2 = norm.rvs(loc = 2, scale = 0.3, size=1000)      # normal de media 2 y sdv 0.3

plt.plot(y)
plt.plot(y2)
plt.grid()
plt.show()

# ¿Cómo hacemos para ajustar una distribución?

plt.hist(y,bins=20)
plt.hist(y2,bins=20)
plt.grid()
plt.show()

[media, varia] = norm.fit(y2)

x2 = np.linspace(y2.min(),y2.max(),100)
pdf2 = norm.pdf(x2, loc = media, scale = varia)

plt.hist(y2, bins = 20, normed = True)
plt.plot(x2,pdf2)
plt.grid()
plt.show()
